from typing import TYPE_CHECKING

from aiosend.types import Balance

from .base import CryptoPayMethod

if TYPE_CHECKING:
    import aiosend


class GetBalance:
    """getBalance method."""

    class GetBalanceMethod(CryptoPayMethod[list[Balance]]):
        __return_type__ = list[Balance]
        __method__ = "getBalance"

    async def get_balance(self: "aiosend.CryptoPay") -> list[Balance]:
        """
        getBalance method.

        Use this method to get balances of your app.
        Requires no parameters.
        Returns array of :class:`Balance`.

        Source: https://help.crypt.bot/crypto-pay-api#getBalance

        :return: List of :class:`Balance` objects.
        """
        return await self(self.GetBalanceMethod(**locals()))



from typing import TYPE_CHECKING

from aiosend.enums import InvoiceStatus

if TYPE_CHECKING:
    import aiosend


class DeleteAllInvoices:
    """Delete all invoices."""

    async def delete_all_invoices(
        self: "aiosend.CryptoPay",
    ) -> None:
        """
        Delete all invoices.

        Wrapper for :class:`aiosend.CryptoPay.get_invoices`
        and :class:`aiosend.CryptoPay.delete_invoice`

        Use this method to delete all existing
        invoices created by your app.

        :return:
        """
        invoices = await self.get_invoices(status=InvoiceStatus.ACTIVE)
        for invoice in invoices:
            await invoice.delete()



from abc import ABC
from collections.abc import Generator
from typing import Generic, TypeAlias, TypeVar

from pydantic import BaseModel
from typing_extensions import Self

from aiosend import CryptoPay

_CryptoPayType = TypeVar(
    "_CryptoPayType",
    bound=CryptoPayObject | list | bool,
)
_T = TypeVar("_T")
SerList: TypeAlias = list[_T]

class CryptoPayObject(BaseModel, ABC):
    _client: CryptoPay
    def __await__(self) -> Generator[None, None, Self]: ...

class Error(BaseModel):
    code: int
    name: str

class ItemsList(BaseModel, Generic[_CryptoPayType]):
    items: _CryptoPayType

class Response(BaseModel, Generic[_CryptoPayType]):
    ok: bool
    result: _CryptoPayType | ItemsList[_CryptoPayType] | None = None
    error: Error | None = None



from abc import ABC, abstractmethod
from types import TracebackType
from typing import TYPE_CHECKING, cast

from pydantic import ValidationError

from aiosend.exceptions import APIError, DeserializationError
from aiosend.methods import (
    CryptoPayMethod,
)
from aiosend.types import (
    ItemsList,
    Response,
    _CryptoPayType,
)

if TYPE_CHECKING:
    import aiosend
    from aiosend.client import APIServer
    from aiosend.types import Error


class BaseSession(ABC):
    """
    Abstract session class.

    If you want to implement your own session class,
    you should inherit this class.
    """

    def __init__(self, api_server: "APIServer") -> None:
        self.api_server = api_server

    @abstractmethod
    async def request(
        self,
        token: str,
        client: "aiosend.CryptoPay",
        method: "CryptoPayMethod[_CryptoPayType]",
    ) -> "_CryptoPayType":
        """Make http request."""

    @abstractmethod
    async def close(self) -> None:
        """Close http session."""

    def _check_response(
        self,
        client: "aiosend.CryptoPay",
        method: CryptoPayMethod[_CryptoPayType],
        content: str,
    ) -> Response[_CryptoPayType]:
        try:
            response = Response[method.__return_type__].model_validate_json(  # type: ignore[name-defined]
                content,
                context={"client": client},
            )
        except ValidationError as e:
            raise DeserializationError(
                method,
                "Failed to deserialize object",
            ) from e

        if not response.ok:
            error = cast("Error", response.error)
            raise APIError(method, error)
        if isinstance(response.result, ItemsList):
            response.result = response.result.items
        return response

    async def __aenter__(self) -> "BaseSession":
        """Enter async context manager."""
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Exit async context manager."""
        await self.close()



import asyncio
import functools
import inspect
import sys
from typing import TYPE_CHECKING

from aiosend.client import CryptoPay
from aiosend.types import CryptoPayObject

if TYPE_CHECKING:
    from collections.abc import Awaitable


def async_to_sync(obj: object, name: str) -> None:
    """Set asyncio event loop if it's not running."""
    method = getattr(obj, name)

    @functools.wraps(method)
    def sync_wrapper(
        *args: object,
        **kwargs: object,
    ) -> object:
        coro: Awaitable = method(*args, **kwargs)
        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
        if loop.is_running():
            return coro
        return loop.run_until_complete(coro)

    setattr(obj, name, sync_wrapper)


def syncify(obj: object) -> None:
    """Add decorators to all public methods of the object."""
    for name in dir(obj):
        if not name.startswith("_") and inspect.iscoroutinefunction(
            getattr(obj, name),
        ):
            async_to_sync(obj, name)


if sys.platform == "win32":
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

for base in (*CryptoPay.__bases__, *CryptoPayObject.__subclasses__()):
    syncify(base)



from typing import TYPE_CHECKING

from aiosend.enums import Asset
from aiosend.types import Check

from .base import CryptoPayMethod

if TYPE_CHECKING:
    import aiosend


class CreateCheck:
    """createCheck method."""

    class CreateCheckMethod(CryptoPayMethod[Check]):
        __return_type__ = Check
        __method__ = "createCheck"

        asset: str
        amount: float
        pin_to_user_id: int | None
        pin_to_username: str | None

    async def create_check(
        self: "aiosend.CryptoPay",
        amount: float,
        asset: Asset,
        pin_to_user_id: int | None = None,
        pin_to_username: str | None = None,
    ) -> Check:
        """
        createCheck method.

        Use this method to create a new check.
        On success, returns an object of the created :class:`Check`.

        Source: https://help.crypt.bot/crypto-pay-api#createCheck

        :param amount: Amount of the check in float.
        :param asset: Cryptocurrency alphabetic code.
        :param pin_to_user_id: ID of the user who will be able to activate the check, defaults to None
        :param pin_to_username: A user with the specified username will be able to activate the check, defaults to None
        :return: :class:`cryptopay.types.Check` object
        """
        return await self(self.CreateCheckMethod(**locals()))



from threading import Thread


class PropagatingThread(Thread):
    """
    A custom thread class that propagates exceptions to the parent thread.

    This class extends the Thread class and overrides the run() method to catch
    and propagate any exceptions that occur during the thread's execution.
    """

    def run(self) -> None:
        """
        Override the run method of the Thread class.

        This method catches any exceptions that occur
        during the thread's execution.
        """
        self._exc: Exception | None = None
        try:
            self.ret = self._target(*self._args, **self._kwargs)  # type: ignore[attr-defined]
        except Exception as e:  # noqa: BLE001
            self._exc = e

    def join(self, timeout: float | None = None) -> None:
        """
        Override the join method of the Thread class.

        This method joins the thread and propagates any exceptions
        that occurred during the thread's execution.
        """
        super().join(timeout=timeout)
        if self._exc is not None:
            raise self._exc



from enum import Enum
from typing import Literal


class CurrencyType(str, Enum):
    """Type of currency."""

    CRYPTO = "crypto"
    FIAT = "fiat"


LiteralCurrencyType = Literal[
    "crypto",
    "fiat",
]



from datetime import datetime

from aiosend.client.client import (  # type: ignore[attr-defined]
    _str,
    bool,  # noqa: A004
    str,  # noqa: A004
)
from aiosend.enums import Asset, CheckStatus, LiteralFiat

from .base import CryptoPayObject

class Check(CryptoPayObject):
    check_id: int
    hash: _str
    asset: Asset | _str
    amount: float
    bot_check_url: _str
    status: CheckStatus | _str
    created_at: datetime
    activated_at: datetime | None = None

    def delete(self) -> bool: ...
    @property
    def qr(self) -> _str: ...
    def get_image(self, fiat: LiteralFiat | _str) -> str: ...



from aiosend.enums import Asset, Fiat

from .base import CryptoPayObject


class Currency(CryptoPayObject):
    """
    Currency object.

    This object represents an `Crypto Pay` currency.
    """

    is_blockchain: bool
    """True, if the currency is blockchain currency."""
    is_stablecoin: bool
    """True, if the currency is stablecoin."""
    is_fiat: bool
    """True, if the currency is fiat."""
    name: str
    """Currency name."""
    code: Fiat | Asset | str
    """Currency code."""
    url: str | None = None
    """Currency url."""
    decimals: int
    """Currency decimals."""



from datetime import datetime

from aiosend.enums import Asset, CheckStatus, LiteralFiat

from .base import CryptoPayObject


class Check(CryptoPayObject):
    """
    Check object.

    Source: https://help.crypt.bot/crypto-pay-api#Check
    """

    check_id: int
    """Unique ID for this check."""
    hash: str
    """Hash of the check."""
    asset: Asset | str
    """Cryptocurrency alphabetic code. Currently, can be “USDT”, “TON”, “BTC”, “ETH”, “LTC”, “BNB”, “TRX” and “USDC” (and “JET” for testnet)."""
    amount: float
    """Amount of the check in float."""
    bot_check_url: str
    """URL should be provided to the user to activate the check."""
    status: CheckStatus | str
    """Status of the check, can be “active” or “activated”."""
    created_at: datetime
    """Date the check was created in ISO 8601 format."""
    activated_at: datetime | None = None
    """Date the check was activated in ISO 8601 format."""

    async def delete(self) -> bool:
        """
        Shortcut for method :class:`aiosend.CryptoPay.delete_check`.

        Use this method to delete check created by your app.
        Returns :code:`True` on success.

        Source: https://help.crypt.bot/crypto-pay-api#deleteCheck

        :return: :code:`True` on success.
        """
        return await self._client.delete_check(self.check_id)

    @property
    def qr(self) -> str:
        """
        Get check qr code.

        :return: check qr code.
        """
        return self._client.session.api_server.get_qr(self.bot_check_url)

    async def get_image(self, fiat: LiteralFiat | str) -> str:
        """
        Get check preview image.

        :return: check image.
        """
        fiat_amount = await self._client.exchange(
            self.amount,
            self.asset,
            fiat,
        )
        return self._client.session.api_server.get_image(
            asset=self.asset,
            asset_amount=self.amount,
            fiat=fiat,
            fiat_amount=fiat_amount,
            main="asset",
        )



from datetime import datetime

from .base import CryptoPayObject


class AppStats(CryptoPayObject):
    """
    AppStats object.

    Source: http://help.crypt.bot/crypto-pay-api#wnPA
    """

    volume: float
    """Total volume of paid invoices in USD."""
    conversion: float
    """Conversion of all created invoices."""
    unique_users_count: int
    """The unique number of users who have paid the invoice."""
    created_invoice_count: int
    """Total created invoice count."""
    paid_invoice_count: int
    """Total paid invoice count."""
    start_at: datetime
    """The date on which the statistics calculation was started in ISO 8601 format."""
    end_at: datetime
    """The date on which the statistics calculation was ended in ISO 8601 format."""



from enum import Enum
from typing import Literal


class InvoiceStatus(str, Enum):
    """Status of invoice."""

    ACTIVE = "active"
    PAID = "paid"
    EXPIRED = "expired"


LiteralInvoiceStatus = Literal[
    "active",
    "paid",
    "expired",
]



from enum import Enum
from typing import Literal


class PaidBtnName(str, Enum):
    """
    Paid button name.

    Label of the button which will be presented
    to a user after the invoice is paid.
    """

    VIEWITEM = "viewItem"
    OPENCHANNEL = "openChannel"
    OPENBOT = "openBot"
    CALLBACK = "callback"


LiteralPaidBtnName = Literal[
    "viewItem",
    "openChannel",
    "openBot",
    "callback",
]



from aiosend.enums import Asset, Fiat

from .base import CryptoPayObject


class ExchangeRate(CryptoPayObject):
    """
    ExchangeRate object.

    Source: https://help.crypt.bot/crypto-pay-api#ExchangeRate
    """

    is_valid: bool
    """True, if the received rate is up-to-date."""
    is_crypto: bool
    """True, if the source is the cryptocurrency."""
    is_fiat: bool
    """True, if the source is the fiat currency."""
    source: Asset | Fiat | str
    """Currency alphabetic code. Currently, can be “USDT”, “TON”, “BTC”, “ETH”, “LTC”, “BNB”, “TRX” and “USDC”."""
    target: Fiat | str
    """Fiat currency code. Currently, can be “USD”, “EUR”, “RUB”, “BYN”, “UAH”, “GBP”, “CNY”, “KZT”, “UZS”, “GEL”, “TRY”, “AMD”, “THB”, “INR”, “BRL”, “IDR”, “AZN”, “AED”, “PLN” and “ILS"."""
    rate: float
    """The current rate of the source asset valued in the target currency."""



from typing import TYPE_CHECKING

from aiosend.exceptions import CryptoPayError

if TYPE_CHECKING:
    import aiosend
    from aiosend.enums import Asset, Fiat


class Exchange:
    """Exchange."""

    async def exchange(
        self: "aiosend.CryptoPay",
        amount: float,
        source: "Asset | Fiat | str",
        target: "Asset | Fiat | str",
    ) -> float:
        """
        Exchange currency.

        Wrapper for :class:`aiosend.CryptoPay.get_exchange_rates`.

        Use this method to get the equivalent amount in
        the target currency for the source currency.

        :return: :class:`float` on success.
        :raise: :class:`CryptoPayError` if there is no such exchange rate.
        """
        ex_rates = await self.get_exchange_rates()
        for ex_rate in ex_rates:
            if ex_rate.source == source and ex_rate.target == target:
                return ex_rate.rate * amount
            if ex_rate.source == target and ex_rate.target == source:
                return amount / ex_rate.rate

        msg = f"Exchange rate for {source} => {target} not found"
        raise CryptoPayError(msg)



from typing import TYPE_CHECKING

from pydantic import field_serializer

from aiosend.enums import Asset, Fiat, InvoiceStatus
from aiosend.types import Invoice, SerList

from .base import CryptoPayMethod

if TYPE_CHECKING:
    import aiosend


class GetInvoices:
    """getInvoices method."""

    class GetInvoicesMethod(CryptoPayMethod[list[Invoice]]):
        __return_type__ = list[Invoice]
        __method__ = "getInvoices"

        asset: str | None
        fiat: str | None
        invoice_ids: SerList[int] | None
        status: InvoiceStatus | None
        offset: int | None
        count: int | None

        @field_serializer("invoice_ids")
        def serialize_ids(self, value: list[int] | None) -> str | None:
            """Serialize list of invoice ids."""
            if value is not None:
                return ",".join(map(str, value))
            return value

    async def get_invoices(
        self: "aiosend.CryptoPay",
        asset: Asset | None = None,
        fiat: Fiat | None = None,
        invoice_ids: list[int] | None = None,
        status: InvoiceStatus | None = None,
        offset: int | None = None,
        count: int | None = None,
    ) -> list[Invoice]:
        """
        getInvoices method.

        Use this method to get invoices created by your app.
        On success, returns array of :class:`Invoice`.

        Source: https://help.crypt.bot/crypto-pay-api#getInvoices

        :param asset: *Optional*. Cryptocurrency alphabetic code.
        :param fiat: *Optional*. Fiat currency code.
        :param invoice_ids: *Optional*. List of invoice IDs separated by comma.
        :param status: *Optional*. Status of invoices to be returned. Available statuses: “active” and “paid”.
        :param offset: *Optional*. Offset needed to return a specific subset of invoices. Defaults to 0.
        :param count: *Optional*. Number of invoices to be returned. Values between 1-1000 are accepted. Defaults to 100.
        :return: List of :class:`Invoice` objects.
        """
        return await self(self.GetInvoicesMethod(**locals()))



import asyncio
import inspect
from dataclasses import dataclass
from functools import partial
from typing import TYPE_CHECKING, cast

from aiosend import loggers
from aiosend.enums import InvoiceStatus
from aiosend.exceptions import CryptoPayError

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Any

    import aiosend
    from aiosend.types import Invoice

    Handler = Callable[..., Any]


@dataclass(slots=True)
class PollingConfig:
    """Polling configuration."""

    timeout: int = 300
    """Timeout in seconds."""
    delay: int = 2
    """Time to wait between the requests in seconds."""


@dataclass(slots=True)
class PollingTask:
    """
    Wrapper for an Invoice.

    This class is used to make a polling task.
    """

    invoice: "Invoice"
    """Invoice object."""
    timeout: int
    """Remaining time for checking the invoice status."""
    data: "dict[str, Any]"
    """Additional payload"""


class PollingManager:
    """
    Polling manager class.

    This class is used to handle payments
    via invoices polling method.
    """

    def __init__(
        self,
        config: PollingConfig,
    ) -> None:
        self._timeout = config.timeout
        self._delay = config.delay
        self._tasks: dict[int, PollingTask] = {}
        self._handler: Handler | None = None
        self._exp_handler: Handler | None = None

    def polling_handler(self) -> "Callable[[Handler], Handler]":
        """
        Register a handler for polling invoice updates.

        Decorator for handler function.

        :return: handler function.
        """

        def wrapper(handler: "Handler") -> "Handler":
            if self._handler is not None:
                msg = "Polling handler has already been declared"
                raise CryptoPayError(msg)
            self._handler = handler
            return handler

        return wrapper

    def expired_handler(self) -> "Callable[[Handler], Handler]":
        """
        Register a handler for timed out invoices.

        Decorator for handler function.

        :return: handler function.
        """

        def wrapper(handler: "Handler") -> "Handler":
            if self._exp_handler is not None:
                msg = "Expired handler has already been declared"
                raise CryptoPayError(msg)
            self._exp_handler = handler
            return handler

        return wrapper

    def _add_invoice(
        self,
        invoice: "Invoice",
        data: "dict[str, Any]",
    ) -> None:
        if self._handler is None:
            msg = "Polling handler hasn't been declared"
            raise CryptoPayError(msg)
        self._tasks[invoice.invoice_id] = PollingTask(
            invoice=invoice,
            timeout=self._timeout,
            data=data,
        )

    @staticmethod
    async def __call_handler(
        handler: "Handler",
        invoice: "Invoice",
        data: "dict[str, Any]",
    ) -> None:
        spec = inspect.getfullargspec(handler)
        is_async = inspect.iscoroutinefunction(handler)
        handler = partial(
            handler,
            invoice,
            **data
            if spec.varkw is not None
            else {k: v for k, v in data.items() if k in spec.args},
        )
        if is_async:
            await handler()
        else:
            handler()

    async def __process_invoice(
        self,
        invoice: "Invoice",
    ) -> None:
        self._tasks[invoice.invoice_id].timeout -= self._delay
        if invoice.status == InvoiceStatus.PAID:
            await self.__call_handler(
                cast("Handler", self._handler),
                invoice,
                self._tasks[invoice.invoice_id].data,
            )
            loggers.polling.info(
                "Invoice id=%d has been paid",
                invoice.invoice_id,
            )
        if (
            self._tasks[invoice.invoice_id].timeout <= 0
            or invoice.status == InvoiceStatus.EXPIRED
        ):
            if self._exp_handler is not None:
                await self.__call_handler(
                    cast("Handler", self._exp_handler),
                    invoice,
                    self._tasks[invoice.invoice_id].data,
                )
            loggers.polling.info(
                "Invoice id=%d timed out",
                invoice.invoice_id,
            )
        if (
            invoice.status in (InvoiceStatus.PAID, InvoiceStatus.EXPIRED)
            or self._tasks[invoice.invoice_id].timeout <= 0
        ):
            del self._tasks[invoice.invoice_id]

    async def start_polling(
        self: "aiosend.CryptoPay",
        parallel: "Callable[[], Any] | None" = None,
    ) -> None:
        """
        Run invoices polling.

        :param parallel: function to run in background.
        :raises CryptoPayError: when polling handler is not declared.
        """
        if parallel is not None:
            loop = asyncio.get_event_loop()
            loop.run_in_executor(None, parallel)
        if self._handler is None:
            msg = (
                "Polling handler hasn't been declared, example:\n"
                "from aiosend import CryptoPay\n"
                'cp = CryptoPay("TOKEN")\n'
                "@cp.polling_handler()\n"
                "async def handler(invoice): ..."
            )
            raise CryptoPayError(msg)
        loggers.polling.info("Start polling")
        while True:
            await asyncio.sleep(self._delay)
            if not self._tasks:
                continue
            invoices = await self.get_invoices(
                invoice_ids=list(self._tasks),
            )
            for invoice in invoices:
                await self.__process_invoice(invoice)
            loggers.polling.debug(
                "Tasks left: %s Waiting %d seconds...",
                len(self._tasks),
                self._delay,
            )



from builtins import bool as _bool
from builtins import float as _float
from builtins import list as _list
from builtins import str as _str
from collections.abc import Awaitable, Callable, Generator
from datetime import datetime
from types import NoneType as _NoneType
from typing import Any, TypeVar

from typing_extensions import Self

from aiosend.client import APIServer
from aiosend.enums import (
    Asset,
    CheckStatus,
    CurrencyType,
    Fiat,
    InvoiceStatus,
    LiteralAsset,
    LiteralCheckStatus,
    LiteralCurrencyType,
    LiteralFiat,
    LiteralInvoiceStatus,
    LiteralPaidBtnName,
    PaidBtnName,
)
from aiosend.methods import CryptoPayMethod
from aiosend.polling import PollingConfig
from aiosend.polling.manager import Handler, PollingTask
from aiosend.types import (
    App,
    AppStats,
    Balance,
    Check,
    Currency,
    ExchangeRate,
    Invoice,
    Transfer,
    _CryptoPayType,
)
from aiosend.webhook import (
    _APP,
    WebhookManager,
)

from .session import BaseSession

_T = TypeVar("_T")

# These classes are needed for syncronous type hinting.
# Stub file annotates the methods as syncronous, but
# while using this lib asynchronously, the actual return type
# is a coroutine that returns an annotated type, so these classes
# will annotate awaited object as return type.

class list(_list[_T]):  # noqa: A001, N801
    def __await__(self) -> Generator[None, None, Self]: ...

class bool(_bool):  # type: ignore[misc]  # noqa: A001, N801
    def __await__(self) -> Generator[None, None, Self]: ...

class str(_str):  # noqa: A001, N801
    def __await__(self) -> Generator[None, None, Self]: ...

class float(_float):  # noqa: A001, N801
    def __await__(self) -> Generator[None, None, Self]: ...

class NoneType(_NoneType):  # type: ignore[misc, valid-type]
    def __await__(self) -> Generator[None, None, None]: ...

class CryptoPay:
    _token: _str
    session: BaseSession
    _timeout: int
    _delay: int
    _tasks: dict[int, PollingTask]
    _handler: Handler | None
    _exp_handler: Handler | None
    _webhook_manager: WebhookManager

    def __init__(
        self,
        token: _str,
        api_server: APIServer = ...,
        session: type[BaseSession] = ...,
        manager: WebhookManager[_APP] | None = None,
        polling_config: PollingConfig | None = None,
    ) -> None: ...
    async def __call__(
        self,
        method: CryptoPayMethod[_CryptoPayType],
    ) -> _CryptoPayType: ...
    def get_me(self) -> App: ...
    def create_invoice(
        self,
        amount: _float,
        asset: Asset | LiteralAsset | _str | None = None,
        *,
        currency_type: CurrencyType | LiteralCurrencyType | _str | None = None,
        fiat: Fiat | LiteralFiat | _str | None = None,
        accepted_assets: _list[Asset | LiteralAsset | _str] | None = None,
        description: _str | None = None,
        hidden_message: _str | None = None,
        paid_btn_name: PaidBtnName | LiteralPaidBtnName | _str | None = None,
        paid_btn_url: _str | None = None,
        payload: _str | None = None,
        allow_comments: bool | None = None,
        allow_anonymous: bool | None = None,
        expires_in: int | None = None,
    ) -> Invoice: ...
    def delete_invoice(
        self,
        invoice_id: int,
    ) -> bool: ...
    def create_check(
        self,
        amount: _float,
        asset: Asset | LiteralAsset | _str,
        pin_to_user_id: int | None = None,
        pin_to_username: _str | None = None,
    ) -> Check: ...
    def delete_check(
        self,
        check_id: int,
    ) -> bool: ...
    def transfer(
        self,
        user_id: int,
        asset: Asset | LiteralAsset | _str,
        amount: _float,
        spend_id: _str | None = None,
        comment: _str | None = None,
        disable_send_notification: bool | None = None,
    ) -> Transfer: ...
    def get_invoices(
        self,
        asset: Asset | LiteralAsset | _str | None = None,
        fiat: Fiat | LiteralFiat | _str | None = None,
        invoice_ids: _list[int] | None = None,
        status: InvoiceStatus | LiteralInvoiceStatus | _str | None = None,
        offset: int | None = None,
        count: int | None = None,
    ) -> list[Invoice]: ...
    def get_checks(
        self,
        asset: Asset | LiteralAsset | _str | None = None,
        check_ids: _list[int] | None = None,
        status: CheckStatus | LiteralCheckStatus | _str | None = None,
        offset: int | None = None,
        count: int | None = None,
    ) -> list[Check]: ...
    def get_transfers(
        self,
        asset: Asset | LiteralAsset | _str | None = None,
        transfer_ids: _list[int] | None = None,
        spend_id: _str | None = None,
        offset: int | None = None,
        count: int | None = None,
    ) -> list[Transfer]: ...
    def get_balance(self) -> list[Balance]: ...
    def get_exchange_rates(self) -> list[ExchangeRate]: ...
    def get_currencies(self) -> list[Currency]: ...
    def get_stats(
        self,
        start_at: datetime | None = None,
        end_at: datetime | None = None,
    ) -> AppStats: ...
    def delete_all_checks(self) -> NoneType: ...
    def delete_all_invoices(self) -> NoneType: ...
    def exchange(
        self,
        amount: _float,
        source: Asset | LiteralAsset | Fiat | LiteralFiat | _str,
        target: Asset | LiteralAsset | Fiat | LiteralFiat | _str,
    ) -> float: ...
    def get_balance_by_asset(
        self,
        asset: Asset | LiteralAsset | _str,
    ) -> float: ...
    def polling_handler(self) -> Callable[[Handler], Handler]: ...
    def expired_handler(self) -> Callable[[Handler], Handler]: ...
    def webhook_handler(
        self,
        app: _APP,
        path: _str,
    ) -> Callable[
        [Callable[[Invoice], Awaitable]],
        Callable[[Invoice], Awaitable],
    ]: ...
    def feed_update(
        self,
        handler: Callable[[Invoice], Awaitable],
        body: dict[_str, Any],
        headers: dict[_str, _str],
    ) -> NoneType: ...
    async def __process_invoice(
        self,
        invoice: Invoice,
    ) -> NoneType: ...
    def _add_invoice(
        self,
        invoice: Invoice,
        data: dict[_str, Any],
    ) -> NoneType: ...
    def start_polling(
        self,
        parallel: Callable[[], Any] | None = None,
    ) -> NoneType: ...
    def get_invoice(
        self,
        invoice: int | Invoice,
    ) -> Invoice | NoneType: ...



from typing import TYPE_CHECKING

from aiosend.enums import CheckStatus

if TYPE_CHECKING:
    import aiosend


class DeleteAllChecks:
    """Delete all checks."""

    async def delete_all_checks(
        self: "aiosend.CryptoPay",
    ) -> None:
        """
        Delete all checks.

        Wrapper for :class:`aiosend.CryptoPay.get_checks`
        and :class:`aiosend.CryptoPay.delete_check`

        Use this method to delete all existing
        checks created by your app.

        :return:
        """
        checks = await self.get_checks(status=CheckStatus.ACTIVE)
        for check in checks:
            await check.delete()



from typing import TYPE_CHECKING

from aiosend.exceptions import CryptoPayError

if TYPE_CHECKING:
    import aiosend
    from aiosend.enums import Asset


class GetBalanceByAsset:
    """Get balance by Asset."""

    async def get_balance_by_asset(
        self: "aiosend.CryptoPay",
        asset: "Asset | str",
    ) -> float:
        """
        Get the balance of a specific asset.

        Wrapper for :class:`aiosend.CryptoPay.get_balance`.

        Use this method to get total avaliable
        amount in float of a specific asset.

        :return: :class:`float` on success.
        :raise: :class:`CryptoPayError` if there is no such asset.
        """
        balances = await self.get_balance()
        for balance in balances:
            if balance.currency_code == asset:
                return balance.available
        msg = f"Balance for {asset} not found"
        raise CryptoPayError(msg)



from .base import CryptoPayMethod
from .create_check import CreateCheck
from .create_invoice import CreateInvoice
from .delete_check import DeleteCheck
from .delete_invoice import DeleteInvoice
from .get_balance import GetBalance
from .get_checks import GetChecks
from .get_currencies import GetCurrencies
from .get_exchange_rates import GetExchangeRates
from .get_invoices import GetInvoices
from .get_me import GetMe
from .get_stats import GetStats
from .get_transfers import GetTransfers
from .transfer import Transfer


class Methods(
    CreateCheck,
    CreateInvoice,
    DeleteCheck,
    DeleteInvoice,
    GetBalance,
    GetChecks,
    GetCurrencies,
    GetExchangeRates,
    GetInvoices,
    GetMe,
    GetStats,
    GetTransfers,
    Transfer,
):
    pass


__all__ = (
    "CreateCheck",
    "CreateInvoice",
    "CryptoPayMethod",
    "DeleteCheck",
    "DeleteInvoice",
    "GetBalance",
    "GetChecks",
    "GetCurrencies",
    "GetExchangeRates",
    "GetInvoices",
    "GetMe",
    "GetStats",
    "GetTransfers",
    "Transfer",
)



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from aiosend.methods import CryptoPayMethod
    from aiosend.types import Error


class CryptoPayError(Exception):
    """Base class for exceptions in this module."""


class APIError(CryptoPayError):
    """Exception for API errors."""

    def __init__(
        self,
        method: "CryptoPayMethod",
        error: "Error",
    ) -> None:
        self.method = method
        self.error = error

    def __str__(self) -> str:
        """Return a string representation of the exception."""
        return (
            f"[{self.error.code}] /{self.method.__method__}: "
            f"{self.error.name} {self.error.model_extra or ''}"
        )


class WrongNetworkError(CryptoPayError):
    """Exception raised when the token is served by different network."""

    def __init__(self, message: str) -> None:
        self.message = message


class DeserializationError(CryptoPayError):
    """Exception for deserialization errors."""

    def __init__(self, method: "CryptoPayMethod", message: str) -> None:
        self.method = method
        self.message = message

    def __str__(self) -> str:
        """Return a string representation of the exception."""
        return f"/{self.method.__method__} {self.message}"


class MethodValuesError(CryptoPayError):
    """Exception raised when method values are invalid."""

    def __init__(self, message: str) -> None:
        self.message = message



from typing import TYPE_CHECKING

from aiosend.types import ExchangeRate

from .base import CryptoPayMethod

if TYPE_CHECKING:
    import aiosend


class GetExchangeRates:
    """getExchangeRates method."""

    class GetExchangeRates(CryptoPayMethod[list[ExchangeRate]]):
        __return_type__ = list[ExchangeRate]
        __method__ = "getExchangeRates"

    async def get_exchange_rates(
        self: "aiosend.CryptoPay",
    ) -> list[ExchangeRate]:
        """
        getExchangeRates method.

        Use this method to get exchange rates of supported currencies.
        Requires no parameters.
        Returns array of :class:`ExchangeRate`.

        Source: https://help.crypt.bot/crypto-pay-api#getExchangeRates

        :return: List of :class:`ExchangeRate` objects.
        """
        return await self(self.GetExchangeRates(**locals()))



from datetime import datetime
from typing import Literal

from aiosend.enums import Asset

from .base import CryptoPayObject


class Transfer(CryptoPayObject):
    """
    Transfer object.

    Source: https://help.crypt.bot/crypto-pay-api#Transfer
    """

    transfer_id: int
    """Unique ID for this transfer."""
    spend_id: str | None = None # not returned by API on /transfer
    """Unique UTF-8 string."""
    user_id: int
    """Telegram user ID the transfer was sent to."""
    asset: Asset | str
    """Cryptocurrency alphabetic code. Currently, can be “USDT”, “TON”, “BTC”, “ETH”, “LTC”, “BNB”, “TRX” and “USDC” (and “JET” for testnet)."""
    amount: float
    """Amount of the transfer in float."""
    status: Literal["completed"]
    """Status of the transfer, can only be “completed”."""
    completed_at: datetime
    """Date the transfer was completed in ISO 8601 format."""
    comment: str | None = None
    """*Optional*. Comment for this transfer."""



import secrets
from typing import TYPE_CHECKING

from pydantic import field_validator

from aiosend.enums import Asset
from aiosend.types import Transfer as TransferType

from .base import CryptoPayMethod

if TYPE_CHECKING:
    import aiosend


class Transfer:
    """transfer method."""

    class TransferMethod(CryptoPayMethod[TransferType]):
        __return_type__ = TransferType
        __method__ = "transfer"

        user_id: int
        asset: str
        amount: float
        spend_id: str | None
        comment: str | None
        disable_send_notification: bool | None

        @field_validator("spend_id", mode="before")
        @classmethod
        def generate_spend_id(cls, v: str | None) -> str:
            """Generate spend_id if not present."""
            return v or secrets.token_hex()

    async def transfer(
        self: "aiosend.CryptoPay",
        user_id: int,
        asset: Asset,
        amount: float,
        spend_id: str | None = None,
        comment: str | None = None,
        disable_send_notification: bool | None = None,
    ) -> TransferType:
        """
        transfer method.

        Use this method to send coins from your app's balance to a user.
        On success, returns completed :class:`Transfer`.
        This method must first be enabled in the security settings of your app.

        Source: https://help.crypt.bot/crypto-pay-api#transfer

        :param user_id: User ID in Telegram. User must have previously used @CryptoBot (@CryptoTestnetBot for testnet).
        :param asset: Cryptocurrency alphabetic code.
        :param amount: Amount of the transfer in float. The minimum and maximum amount limits for each of the supported assets roughly correspond to 1-25000 USD.
        :param spend_id: Random UTF-8 string unique per transfer for idempotent requests. The same spend_id can be accepted only once from your app. Up to 64 symbols.
        :param comment: *Optional*. Comment for the transfer. Users will see this comment in the notification about the transfer. Up to 1024 symbols.
        :param disable_send_notification: *Optional*. Pass true to not send to the user the notification about the transfer. Defaults to false.
        :return: :class:`Transfer` object.
        """
        return await self(self.TransferMethod(**locals()))



from .aiohttp_manager import AiohttpManager
from .base import _APP, RequestHandler, WebhookManager
from .fastapi_manager import FastAPIManager
from .flask_manager import FlaskManager

__all__ = (
    "_APP",
    "AiohttpManager",
    "FastAPIManager",
    "FlaskManager",
    "RequestHandler",
    "WebhookManager",
)



from typing import TYPE_CHECKING

from aiosend.types import Currency

from .base import CryptoPayMethod

if TYPE_CHECKING:
    import aiosend


class GetCurrencies:
    """getCurrencies method."""

    class GetCurrenciesMethod(CryptoPayMethod[list[Currency]]):
        __return_type__ = list[Currency]
        __method__ = "getCurrencies"

    async def get_currencies(
        self: "aiosend.CryptoPay",
    ) -> list[Currency]:
        """
        getCurrencies method.

        Use this method to get a list of supported currencies.
        Requires no parameters.
        Returns a list of fiat and cryptocurrency alphabetic codes.

        Source: https://help.crypt.bot/crypto-pay-api#getCurrencies

        :return: List of :class:`Currency` objects.
        """
        return await self(self.GetCurrenciesMethod(**locals()))



from datetime import datetime
from typing import cast

from aiosend.enums import (
    Asset,
    CurrencyType,
    Fiat,
    InvoiceStatus,
    PaidBtnName,
)

from .base import CryptoPayObject


class Invoice(CryptoPayObject):
    """
    Invoice object.

    Source: https://help.crypt.bot/crypto-pay-api#Invoice
    """

    invoice_id: int
    """Unique ID for this invoice."""
    hash: str
    """Hash of the invoice."""
    currency_type: CurrencyType | str
    """Type of the price, can be “crypto” or “fiat”."""
    asset: Asset | str | None = None
    """*Optional*. Cryptocurrency code. Available only if the value of the field currency_type is “crypto”. Currently, can be “USDT”, “TON”, “BTC”, “ETH”, “LTC”, “BNB”, “TRX” and “USDC” (and “JET” for testnet)."""
    fiat: Fiat | str | None = None
    """*Optional*. Fiat currency code. Available only if the value of the field currency_type is “fiat”. Currently one of “USD”, “EUR”, “RUB”, “BYN”, “UAH”, “GBP”, “CNY”, “KZT”, “UZS”, “GEL”, “TRY”, “AMD”, “THB”, “INR”, “BRL”, “IDR”, “AZN”, “AED”, “PLN” and “ILS"."""
    amount: float
    """Amount of the invoice for which the invoice was created."""
    paid_asset: Asset | str | None = None
    """*Optional*. Cryptocurrency alphabetic code for which the invoice was paid. Available only if currency_type is “fiat” and status is “paid”."""
    paid_amount: float | None = None
    """*Optional*. Amount of the invoice for which the invoice was paid. Available only if currency_type is “fiat” and status is “paid”."""
    paid_fiat_rate: float | None = None
    """*Optional*. The rate of the paid_asset valued in the fiat currency. Available only if the value of the field currency_type is “fiat” and the value of the field status is “paid”."""
    accepted_assets: list[Asset | str] | None = None
    """*Optional*. List of assets which can be used to pay the invoice. Available only if currency_type is “fiat”. Currently, can be “USDT”, “TON”, “BTC”, “ETH”, “LTC”, “BNB”, “TRX” and “USDC” (and “JET” for testnet)."""
    fee_asset: Asset | str | None = None
    """*Optional*. Asset of service fees charged when the invoice was paid. Available only if status is “paid”."""
    fee_amount: float | None = None
    """*Optional*. Amount of service fees charged when the invoice was paid. Available only if status is “paid”."""
    fee: str | None = None  # deprecated
    """*Optional*. Amount of charged service fees. Available only in the payload of the webhook update (described here for reference)."""
    pay_url: str | None = None  # deprecated
    """*Deprecated*. URL should be provided to the user to pay the invoice (described here for reference)."""
    bot_invoice_url: str
    """URL should be provided to the user to pay the invoice."""
    mini_app_invoice_url: str
    """Use this URL to pay an invoice to the Telegram Mini App version."""
    web_app_invoice_url: str
    """Use this URL to pay an invoice to the Web version of Crypto Bot."""
    description: str | None = None
    """*Optional*. Description for this invoice."""
    status: InvoiceStatus | str | None = None
    """Status of the transfer, can be “active”, “paid” or “expired”."""
    created_at: datetime
    """Date the invoice was created in ISO 8601 format."""
    paid_usd_rate: float | None = None
    """*Optional*. Price of the asset in USD. Available only if status is “paid”."""
    usd_rate: str | None = None  # deprecated
    """*Optional*. Price of the asset in USD. Available only in the Webhook update payload."""
    allow_comments: bool
    """True, if the user can add comment to the payment."""
    allow_anonymous: bool
    """True, if the user can pay the invoice anonymously."""
    expiration_date: datetime | None = None
    """*Optional*. Date the invoice expires in ISO 8601 format."""
    paid_at: datetime | None = None
    """*Optional*. Date the invoice was paid in ISO 8601 format."""
    paid_anonymously: bool | None = None
    """True, if the invoice was paid anonymously."""
    comment: str | None = None
    """*Optional*. Comment to the payment from the user."""
    hidden_message: str | None = None
    """*Optional*. Text of the hidden message for this invoice."""
    payload: str | None = None
    """*Optional*. Previously provided data for this invoice."""
    paid_btn_name: PaidBtnName | str | None = None
    """*Optional*. Label of the button, can be “viewItem”, “openChannel”, “openBot” or “callback”."""
    paid_btn_url: str | None = None
    """*Optional*. URL opened using the button."""

    async def delete(self) -> bool:
        """
        Shortcut for method :class:`aiosend.CryptoPay.delete_invoice`.

        Use this method to delete invoice created by your app.
        Returns :code:`True` on success.

        Source: https://help.crypt.bot/crypto-pay-api#hwjK

        :return: :code:`True` on success.
        """
        return await self._client.delete_invoice(self.invoice_id)

    async def update(self) -> None:
        """
        Shortcut for method :class:`aiosend.CryptoPay.get_invoices`.

        Use this method to update invoice object.

        Source: https://help.crypt.bot/crypto-pay-api#getInvoices

        :return:
        """
        invoice = cast(Invoice, await self._client.get_invoice(self))
        self.__dict__ = invoice.__dict__

    def await_payment(self, **kwargs: object) -> None:
        """
        Send the invoice to the polling manager.

        Use this method to check the status of
        the invoice until the timeout expires.

        :param kwargs: additional payload for the handler.

        :return:
        """
        self._client._add_invoice(self, kwargs)  # noqa: SLF001

    @property
    def qr(self) -> str:
        """
        Get invoice qr code.

        :return: invoice qr code.
        """
        return self._client.session.api_server.get_qr(self.bot_invoice_url)



from datetime import datetime
from typing import TYPE_CHECKING

from aiosend.types import AppStats

from .base import CryptoPayMethod

if TYPE_CHECKING:
    import aiosend


class GetStats:
    """getStats method."""

    class GetStatsMethod(CryptoPayMethod[AppStats]):
        __return_type__ = AppStats
        __method__ = "getStats"

        start_at: datetime | None
        end_at: datetime | None

    async def get_stats(
        self: "aiosend.CryptoPay",
        start_at: datetime | None = None,
        end_at: datetime | None = None,
    ) -> AppStats:
        """
        getStats method.

        Use this method to get app statistics.
        On success, returns :class:`AppStats`.

        Source: https://help.crypt.bot/crypto-pay-api#jvP3

        :param start_at: *Optional*. Date from which start calculating statistics. Defaults is current date minus 24 hours.
        :param end_at: *Optional*. The date on which to finish calculating statistics. Defaults is current date.
        :return: :class:`AppStats` object.
        """
        return await self(self.GetStatsMethod(**locals()))



from .app import App
from .app_stats import AppStats
from .balance import Balance
from .base import (
    CryptoPayObject,
    Error,
    ItemsList,
    Response,
    SerList,
    _CryptoPayType,
)
from .check import Check
from .currency import Currency
from .exchange_rate import ExchangeRate
from .invoice import Invoice
from .transfer import Transfer
from .update import Update

__all__ = (
    "App",
    "AppStats",
    "Balance",
    "Check",
    "CryptoPayObject",
    "Currency",
    "Error",
    "ExchangeRate",
    "Invoice",
    "ItemsList",
    "Response",
    "SerList",
    "Transfer",
    "Update",
    "_CryptoPayType",
)



import functools
import hashlib
import json
from abc import ABC, abstractmethod
from hmac import HMAC
from typing import TYPE_CHECKING, Any, Generic, TypeVar

from aiosend import loggers
from aiosend.types import Update

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable, Mapping

    import aiosend
    from aiosend.types import Invoice

    InvoiceHandler = Callable[[Invoice], Awaitable]
    Handler = Callable[[dict[str, Any], Mapping[str, str]], Awaitable]

_APP = TypeVar("_APP")


class WebhookManager(Generic[_APP], ABC):
    """
    Webhook manager.

    If you want to implement your own webhook
    manager, you must inherit from this class.
    """

    @abstractmethod
    def register_handler(
        self,
        app: _APP,
        path: str,
        handler: "Handler",
    ) -> None:
        """
        Register webhook handler.

        Override this method in your own webhook manager class.
        This method is used for registering webhook handler in your app.
        You should call `handler` whenever an update arrives.

        :param app: Web application.
        :param path: Path to webhook route.
        :param handler: Handler object.
        :return:
        """


class RequestHandler(Generic[_APP]):
    """Updates handler."""

    def __init__(
        self,
        manager: WebhookManager[_APP],
    ) -> None:
        self._webhook_manager = manager

    def _check_signature(
        self: "aiosend.CryptoPay",
        body: "dict[str, Any]",
        headers: dict[str, str],
    ) -> bool:
        """
        Verify the received update and the integrity of the received data.

        Source: https://help.crypt.bot/crypto-pay-api#verifying-webhook-updates

        :param body: unparsed JSON string.
        :param headers: request headers.
        :return: True if the signature is correct, False otherwise.
        """
        secret = hashlib.sha256(self._token.encode()).digest()
        check_string = json.dumps(
            body,
            separators=(",", ":"),
            ensure_ascii=False,
        )
        hmac = HMAC(secret, check_string.encode(), hashlib.sha256).hexdigest()
        return hmac == headers.get("Crypto-Pay-Api-Signature")

    async def feed_update(
        self,
        handler: "InvoiceHandler",
        body: dict[str, Any],
        headers: dict[str, str],
    ) -> None:
        """
        Feed an update to the invoice handler.

        :param body: parsed json body.
        :param headers: request headers.
        """
        update = Update.model_validate(body, context={"client": self})
        if self._check_signature(body, headers):
            await handler(update.payload)
            loggers.webhook.info(
                "Webhook Update id=%d is handled.",
                update.update_id,
            )
        else:
            loggers.webhook.info(
                "Webhook Update id=%d is not handled. Signature is invalid.",
                update.update_id,
            )

    def webhook_handler(
        self,
        app: _APP,
        path: str,
    ) -> "Callable[[InvoiceHandler], InvoiceHandler]":
        """
        Register a handler for webhook invoice updates.

        :param app: web application.
        :param path: webhook path.

        :return: handler function.
        """

        def wrapper(handler: "InvoiceHandler") -> "InvoiceHandler":
            self._webhook_manager.register_handler(
                app,
                path,
                functools.partial(self.feed_update, handler),
            )
            return handler

        return wrapper



from .asset import Asset, LiteralAsset
from .check_status import CheckStatus, LiteralCheckStatus
from .currency_type import CurrencyType, LiteralCurrencyType
from .fiat import Fiat, LiteralFiat
from .invoice_status import InvoiceStatus, LiteralInvoiceStatus
from .paid_btn_name import LiteralPaidBtnName, PaidBtnName
from .update_type import LiteralUpdateType, UpdateType

__all__ = (
    "Asset",
    "CheckStatus",
    "CurrencyType",
    "Fiat",
    "InvoiceStatus",
    "LiteralAsset",
    "LiteralCheckStatus",
    "LiteralCurrencyType",
    "LiteralFiat",
    "LiteralInvoiceStatus",
    "LiteralPaidBtnName",
    "LiteralUpdateType",
    "PaidBtnName",
    "UpdateType",
)



from typing import TYPE_CHECKING

from pydantic import Field

from aiosend.enums import Asset
from aiosend.types import SerList, Transfer

from .base import CryptoPayMethod

if TYPE_CHECKING:
    import aiosend


class GetTransfers:
    """getTransfers method."""

    class GetTransfersMethod(CryptoPayMethod[list[Transfer]]):
        __return_type__ = list[Transfer]
        __method__ = "getTransfers"

        asset: str | None
        transfer_ids: SerList[int] | None
        spend_id: str | None
        offset: int | None
        count: int | None = Field(None, ge=1, le=1000)

    async def get_transfers(
        self: "aiosend.CryptoPay",
        asset: Asset | None = None,
        transfer_ids: list[int] | None = None,
        spend_id: str | None = None,
        offset: int | None = None,
        count: int | None = None,
    ) -> list[Transfer]:
        """
        getTransfers method.

        Use this method to get transfers created by your app.
        On success, returns array of :class:`Transfer`.

        Source: https://help.crypt.bot/crypto-pay-api#getTransfers

        :param asset: *Optional*. Cryptocurrency alphabetic code. Defaults to all currencies.
        :param transfer_ids: *Optional*. List of transfer IDs.
        :param spend_id: *Optional*. Unique UTF-8 transfer string.
        :param offset: *Optional*. Offset needed to return a specific subset of transfers. Defaults to 0.
        :param count: *Optional*. Number of transfers to be returned. Values between 1-1000 are accepted. Defaults to 100.
        :return: List of :class:`Transfer` objects.
        """
        return await self(self.GetTransfersMethod(**locals()))



from abc import ABC
from typing import TYPE_CHECKING, Annotated, Generic, TypeVar

from pydantic import BaseModel, ConfigDict, PrivateAttr
from pydantic.functional_serializers import PlainSerializer

from aiosend.utils import serialize_list

if TYPE_CHECKING:
    from aiosend import CryptoPay

_CryptoPayType = TypeVar(
    "_CryptoPayType",
    bound="CryptoPayObject | list | bool",
)
_T = TypeVar("_T")
SerList = Annotated[list[_T], PlainSerializer(serialize_list, str | None)]


class CryptoPayObject(BaseModel, ABC):
    """Base object class for types."""

    _client: "CryptoPay" = PrivateAttr()

    model_config = ConfigDict(
        extra="allow",
    )

    def model_post_init(self, ctx: dict) -> None:
        """Bind client to CryptoPayObject."""
        self._client = ctx["client"]


class Error(BaseModel):
    """API error model."""

    code: int
    name: str

    model_config = ConfigDict(
        extra="allow",
        frozen=True,
    )


class ItemsList(BaseModel, Generic[_CryptoPayType]):
    """
    Items list.

    This model is used to convert a dictionary with the `items` key to a list.
    """

    items: _CryptoPayType


class Response(BaseModel, Generic[_CryptoPayType]):
    """API response model."""

    ok: bool
    result: _CryptoPayType | ItemsList[_CryptoPayType] | None = None
    error: Error | None = None



from datetime import datetime

from aiosend.client.client import (  # type: ignore[attr-defined]
    NoneType,
    _bool,
    bool,  # noqa: A004
)
from aiosend.enums import (
    Asset,
    CurrencyType,
    Fiat,
    InvoiceStatus,
    PaidBtnName,
)

from .base import CryptoPayObject

class Invoice(CryptoPayObject):
    invoice_id: int
    hash: str
    currency_type: CurrencyType | str
    asset: Asset | str | None = None
    fiat: Fiat | str | None = None
    amount: float
    paid_asset: Asset | str | None = None
    paid_amount: float | None = None
    paid_fiat_rate: float | None = None
    accepted_assets: list[Asset | str] | None = None
    fee_asset: Asset | str | None = None
    fee_amount: float | None = None
    fee: str | None = None
    pay_url: str | None = None
    bot_invoice_url: str
    mini_app_invoice_url: str
    web_app_invoice_url: str
    description: str | None = None
    status: InvoiceStatus | str | None = None
    created_at: datetime
    paid_usd_rate: float | None = None
    usd_rate: str | None = None
    allow_comments: _bool
    allow_anonymous: _bool
    expiration_date: datetime | None = None
    paid_at: datetime | None = None
    paid_anonymously: _bool | None = None
    comment: str | None = None
    hidden_message: str | None = None
    payload: str | None = None
    paid_btn_name: PaidBtnName | str | None = None
    paid_btn_url: str | None = None

    def delete(self) -> bool: ...
    def update(self) -> NoneType: ...
    def await_payment(self, **kwargs: object) -> None: ...
    @property
    def qr(self) -> str: ...



from enum import Enum
from typing import Literal


class Fiat(str, Enum):
    """Fiat currency code."""

    USD = "USD"
    EUR = "EUR"
    RUB = "RUB"
    BYN = "BYN"
    UAH = "UAH"
    GBP = "GBP"
    CNY = "CNY"
    KZT = "KZT"
    UZS = "UZS"
    GEL = "GEL"
    TRY = "TRY"
    AMD = "AMD"
    THB = "THB"
    INR = "INR"
    BRL = "BRL"
    IDR = "IDR"
    AZN = "AZN"
    AED = "AED"
    PLN = "PLN"
    ILS = "ILS"
    KGS = "KGS"
    TJS = "TJS"


LiteralFiat = Literal[
    "USD",
    "EUR",
    "RUB",
    "BYN",
    "UAH",
    "GBP",
    "CNY",
    "KZT",
    "UZS",
    "GEL",
    "TRY",
    "AMD",
    "THB",
    "INR",
    "BRL",
    "IDR",
    "AZN",
    "AED",
    "PLN",
    "ILS",
    "KGS",
    "TJS",
]



from aiosend.client import MAINNET, TESTNET
from aiosend.utils.sync import CryptoPay

from .__meta__ import __api_version__, __version__

__all__ = (
    "MAINNET",
    "TESTNET",
    "CryptoPay",
    "__api_version__",
    "__version__",
)



from aiosend.enums import Asset

from .base import CryptoPayObject


class Balance(CryptoPayObject):
    """
    Balance object.

    Source: https://help.crypt.bot/crypto-pay-api#Balance
    """

    currency_code: Asset | str
    """Cryptocurrency alphabetic code. Currently, can be “USDT”, “TON”, “BTC”, “ETH”, “LTC”, “BNB”, “TRX” and “USDC” (and “JET” for testnet)."""
    available: float
    """Total available amount in float."""
    onhold: float
    """Unavailable amount currently is on hold in float."""



from typing import TYPE_CHECKING

from .base import CryptoPayMethod

if TYPE_CHECKING:
    import aiosend


class DeleteInvoice:
    """deleteInvoice method."""

    class DeleteInvoiceMethod(CryptoPayMethod[bool]):
        __return_type__ = bool
        __method__ = "deleteInvoice"

        invoice_id: int

    async def delete_invoice(
        self: "aiosend.CryptoPay",
        invoice_id: int,
    ) -> bool:
        """
        deleteInvoice method.

        Use this method to delete invoices created by your app.
        Returns :code:`True` on success.

        Source: https://help.crypt.bot/crypto-pay-api#hwjK

        :param invoice_id: Invoice ID to be deleted.
        :return: :code:`True` on success.
        """
        return await self(self.DeleteInvoiceMethod(**locals()))



from enum import Enum
from typing import Literal


class Asset(str, Enum):
    """Cryptocurrency alphabetic code."""

    USDT = "USDT"
    TON = "TON"
    SOL = "SOL"
    GRAM = "GRAM"
    NOT = "NOT"
    HMSTR = "HMSTR"
    CATI = "CATI"
    MY = "MY"
    DOGS = "DOGS"
    BTC = "BTC"
    LTC = "LTC"
    ETH = "ETH"
    BNB = "BNB"
    TRX = "TRX"
    USDC = "USDC"
    JET = "JET"
    SEND = "SEND"
    PEPE = "PEPE"
    WIF = "WIF"
    BONK = "BONK"
    MAJOR = "MAJOR"


LiteralAsset = Literal[
    "USDT",
    "TON",
    "SOL",
    "GRAM",
    "NOT",
    "HMSTR",
    "CATI",
    "MY",
    "DOGS",
    "BTC",
    "LTC",
    "ETH",
    "BNB",
    "TRX",
    "USDC",
    "JET",
    "SEND",
    "PEPE",
    "WIF",
    "BONK",
    "MAJOR",
]



from typing import TYPE_CHECKING

from aiohttp.web import json_response

from .base import WebhookManager

if TYPE_CHECKING:
    from aiohttp.web import Application, Request, Response

    from .base import Handler


class AiohttpManager(WebhookManager["Application"]):
    """
    aiohttp webhook manager.

    Webhook manager based on :class:`aiohttp.web.Application`.
    """

    def register_handler(
        self,
        app: "Application",
        path: str,
        handler: "Handler",
    ) -> None:
        """Register webhook handler."""

        async def handle(request: "Request") -> "Response":
            await handler(
                await request.json(),
                dict(request.headers),
            )
            return json_response({"ok": True})

        app.router.add_post(path, handle)



from typing import TYPE_CHECKING

from pydantic import Field, model_validator

from aiosend.enums import Asset, CurrencyType, Fiat, PaidBtnName
from aiosend.exceptions import MethodValuesError
from aiosend.types import Invoice, SerList

from .base import CryptoPayMethod

if TYPE_CHECKING:
    import aiosend


class CreateInvoice:
    """createInvoice method."""

    class CreateInvoiceMethod(CryptoPayMethod[Invoice]):
        __return_type__ = Invoice
        __method__ = "createInvoice"

        currency_type: str | None
        asset: str | None
        fiat: str | None
        accepted_assets: SerList[Asset] | None
        amount: float
        description: str | None
        hidden_message: str | None
        paid_btn_name: str | None
        paid_btn_url: str | None
        payload: str | None
        allow_comments: bool | None
        allow_anonymous: bool | None
        expires_in: int | None = Field(None, ge=0, le=2678400)

        @model_validator(mode="before")
        @classmethod
        def values_validate(cls, values: dict) -> dict:
            """Validate model values."""
            errors = {
                values["currency_type"] in (None, "crypto")
                and values["asset"]
                is None: 'asset is required if currency_type is "crypto"',
                values["currency_type"] == "fiat"
                and values["fiat"]
                is None: 'fiat is required if currency_type is "fiat"',
                values["currency_type"] != "fiat"
                and values["accepted_assets"]
                is not None: 'accepted_assets avaliable only if currency_type is "fiat"',
                values["paid_btn_name"] is not None
                and values["paid_btn_url"]
                is None: "paid_btn_url required if paid_btn_name is specified.",
            }
            for error, message in errors.items():
                if error:
                    raise MethodValuesError(message)

            return values

    async def create_invoice(
        self: "aiosend.CryptoPay",
        amount: float,
        asset: Asset | None = None,
        *,
        currency_type: CurrencyType | None = None,
        fiat: Fiat | None = None,
        accepted_assets: list[Asset] | None = None,
        description: str | None = None,
        hidden_message: str | None = None,
        paid_btn_name: PaidBtnName | None = None,
        paid_btn_url: str | None = None,
        payload: str | None = None,
        allow_comments: bool | None = None,
        allow_anonymous: bool | None = None,
        expires_in: int | None = None,
    ) -> Invoice:
        """
        createInvoice method.

        Use this method to create a new invoice.
        On success, returns an object of the created :class:`Invoice`.

        Source: https://help.crypt.bot/crypto-pay-api#createInvoice

        :param amount: Amount of the invoice in float.
        :param currency_type: *Optional*. Type of the price, can be “crypto” or “fiat”. Defaults to crypto.
        :param asset: *Optional*. Required if currency_type is “crypto”. Cryptocurrency alphabetic code.
        :param fiat: *Optional*. Required if currency_type is “fiat”. Fiat currency code.
        :param accepted_assets: *Optional*. List of cryptocurrency alphabetic codes. Assets which can be used to pay the invoice. Available only if currency_type is “fiat”
        :param description: *Optional*. Description for the invoice. User will see this description when they pay the invoice. Up to 1024 characters.
        :param hidden_message: *Optional*. Optional. Text of the message which will be presented to a user after the invoice is paid. Up to 2048 characters.
        :param paid_btn_name: *Optional*. Label of the button which will be presented to a user after the invoice is paid.
        :param paid_btn_url: *Optional*. Required if paid_btn_name is specified. URL opened using the button which will be presented to a user after the invoice is paid.
        :param payload: *Optional*. Any data you want to attach to the invoice (for example, user ID, payment ID, ect). Up to 4kb.
        :param allow_comments: *Optional*. Allow a user to add a comment to the payment. Defaults to true.
        :param allow_anonymous: *Optional*. Allow a user to pay the invoice anonymously. Defaults to true.
        :param expires_in: *Optional*. You can set a payment time limit for the invoice in seconds. Values between 1-2678400 are accepted.
        :return: :class:`Invoice` object
        """
        return await self(self.CreateInvoiceMethod(**locals()))



from .delete_all_checks import DeleteAllChecks
from .delete_all_invoices import DeleteAllInvoices
from .exchange import Exchange
from .get_balance_by_asset import GetBalanceByAsset
from .get_invoice import GetInvoice


class Tools(
    DeleteAllChecks,
    DeleteAllInvoices,
    Exchange,
    GetBalanceByAsset,
    GetInvoice,
): ...


__all__ = (
    "DeleteAllChecks",
    "DeleteAllInvoices",
    "Exchange",
    "GetBalanceByAsset",
    "GetInvoice",
)



from abc import ABC
from typing import ClassVar, Generic

from pydantic import BaseModel, ConfigDict

from aiosend.types import _CryptoPayType


class CryptoPayMethod(BaseModel, Generic[_CryptoPayType], ABC):
    """Base `Crypto Pay API` method class."""

    model_config = ConfigDict(
        extra="ignore",
        frozen=True,
    )

    __return_type__: ClassVar[type[_CryptoPayType]]
    __method__: ClassVar[str]



from enum import Enum
from typing import Literal


class CheckStatus(str, Enum):
    """Status of check."""

    ACTIVE = "active"
    ACTIVATED = "activated"


LiteralCheckStatus = Literal[
    "active",
    "activated",
]



from typing import TYPE_CHECKING

from .base import CryptoPayMethod

if TYPE_CHECKING:
    import aiosend


class DeleteCheck:
    """deleteCheck method."""

    class DeleteCheckMethod(CryptoPayMethod[bool]):
        __return_type__ = bool
        __method__ = "deleteCheck"

        check_id: int

    async def delete_check(
        self: "aiosend.CryptoPay",
        check_id: int,
    ) -> bool:
        """
        deleteCheck method.

        Use this method to delete checks created by your app.
        Returns :code:`True` on success.

        Source: https://help.crypt.bot/crypto-pay-api#deleteCheck

        :param check_id: Check ID to be deleted.
        :return: :code:`True` on success.
        """
        return await self(self.DeleteCheckMethod(**locals()))



from .base import CryptoPayObject


class App(CryptoPayObject):
    """
    App object.

    This object represents the `Crypto Pay` app.
    """

    app_id: int
    """Crypto Pay app id."""
    name: str
    """Crypto Pay app name."""
    payment_processing_bot_username: str
    """Telegram username of the payment processing bot."""



from typing import TYPE_CHECKING

from .base import WebhookManager

if TYPE_CHECKING:
    from fastapi import FastAPI

    from .base import Handler


class FastAPIManager(WebhookManager["FastAPI"]):
    """
    FastAPI webhook manager.

    Webhook manager based on :class:`fastapi.FastAPI`.
    """

    def register_handler(
        self,
        app: "FastAPI",
        path: str,
        handler: "Handler",
    ) -> None:
        """Register webhook handler."""
        try:
            from fastapi import Request
        except ModuleNotFoundError as e:
            msg = "fastapi is not installed"
            raise RuntimeError(msg) from e

        @app.post(path)
        async def handle(request: Request) -> dict:
            await handler(
                await request.json(),
                request.headers,
            )
            return {"ok": True}



from typing import TYPE_CHECKING

from aiosend.enums import Asset, CheckStatus
from aiosend.types import Check, SerList

from .base import CryptoPayMethod

if TYPE_CHECKING:
    import aiosend


class GetChecks:
    """getChecks method."""

    class GetChecksMethod(CryptoPayMethod[list[Check]]):
        __return_type__ = list[Check]
        __method__ = "getChecks"

        asset: str | None = None
        check_ids: SerList[int] | None
        status: str | None
        offset: int | None
        count: int | None

    async def get_checks(
        self: "aiosend.CryptoPay",
        asset: Asset | None = None,
        check_ids: list[int] | None = None,
        status: CheckStatus | None = None,
        offset: int | None = None,
        count: int | None = None,
    ) -> list[Check]:
        """
        getChecks method.

        Use this method to get checks created by your app.
        On success, returns array of :class:`Check`.

        Source: https://help.crypt.bot/crypto-pay-api#getChecks

        :param asset: *Optional*. Cryptocurrency alphabetic code.
        :param check_ids: *Optional*. List of check IDs.
        :param status: *Optional*. Status of check to be returned. Available statuses: “active” and “activated”.
        :param offset: *Optional*. Offset needed to return a specific subset of check. Defaults to 0.
        :param count: *Optional*. Number of check to be returned. Values between 1-1000 are accepted. Defaults to 100.
        :return: List of :class:`Check` objects.
        """
        return await self(self.GetChecksMethod(**locals()))



from datetime import datetime

from aiosend.enums import UpdateType

from .base import CryptoPayObject
from .invoice import Invoice


class Update(CryptoPayObject):
    """
    Update object.

    Source: https://help.crypt.bot/crypto-pay-api#webhook-updates
    """

    update_id: int
    update_type: UpdateType
    request_date: datetime
    payload: Invoice



from enum import Enum
from typing import TypeVar

T = TypeVar("T")


def serialize_list(value: list[T] | None) -> str | None:
    """
    List of objects serialization.

    Serialize a list of objects into a
    string of objects separated by commas.
    """
    if value is not None:
        return ",".join(
            [x.value if isinstance(x, Enum) else str(x) for x in value],
        )
    return value



from typing import TYPE_CHECKING

from aiosend.types import App

from .base import CryptoPayMethod

if TYPE_CHECKING:
    import aiosend


class GetMe:
    """getMe method."""

    class GetMeMethod(CryptoPayMethod[App]):
        __return_type__ = App
        __method__ = "getMe"

    async def get_me(self: "aiosend.CryptoPay") -> App:
        """
        getMe method.

        Use this method to test your app's authentication token.
        Requires no parameters.
        On success, returns basic information about an app.

        Source: https://help.crypt.bot/crypto-pay-api#getMe

        :return: :class:`App` object.
        """
        return await self(self.GetMeMethod(**locals()))



from typing import TYPE_CHECKING

from aiosend import loggers
from aiosend.client import MAINNET, TESTNET
from aiosend.exceptions import APIError, WrongNetworkError
from aiosend.methods import Methods
from aiosend.polling import PollingConfig, PollingManager
from aiosend.tools import Tools
from aiosend.utils import PropagatingThread
from aiosend.webhook import AiohttpManager, RequestHandler

from .session import AiohttpSession

if TYPE_CHECKING:
    from aiosend.client import APIServer
    from aiosend.methods import CryptoPayMethod
    from aiosend.types import _CryptoPayType
    from aiosend.webhook import _APP, WebhookManager

    from .session import BaseSession


class CryptoPay(Methods, Tools, RequestHandler, PollingManager):
    """
    Client class providing API methods.

    :param token: Crypto Bot API token
    :param session: HTTP Session
    :param api_server: Crypto Bot API server
    """

    def __init__(
        self,
        token: str,
        api_server: "APIServer" = MAINNET,
        session: "type[BaseSession]" = AiohttpSession,
        manager: "WebhookManager[_APP] | None" = None,
        polling_config: "PollingConfig | None" = None,
    ) -> None:
        self._token = token
        self.session = session(api_server)
        RequestHandler.__init__(self, manager or AiohttpManager())
        PollingManager.__init__(self, polling_config or PollingConfig())
        thread = PropagatingThread(target=self.__auth)
        thread.start()
        thread.join()

    async def __call__(
        self,
        method: "CryptoPayMethod[_CryptoPayType]",
    ) -> "_CryptoPayType":
        """
        Request method.

        Use this method to make an API request.

        :param method: CryptoPayMethod object.
        :return: :class:`CryptoPayType` object.
        """
        async with self.session as session:
            loggers.client.debug("Requesting: %s", method.__method__)
            return await session.request(self._token, self, method)

    def __auth(self) -> None:
        try:
            me = self.get_me()
            loggers.client.info(
                "Authorized as '%s' id=%d on %s",
                me.name,
                me.app_id,
                self.session.api_server.name,
            )
        except APIError:
            current_net = self.session.api_server
            if current_net == MAINNET:
                self.session = self.session.__class__(TESTNET)
            else:
                self.session = self.session.__class__(MAINNET)
            self.get_me()
            net = self.session.api_server
            msg = (
                "Authorization failed. Token is served by the "
                f"{net.name}, you are using {current_net.name}"
            )
            raise WrongNetworkError(msg) from None



import ssl
from typing import TYPE_CHECKING, cast

import certifi
from aiohttp import ClientSession, TCPConnector
from aiohttp.http import SERVER_SOFTWARE

from aiosend.__meta__ import __version__

from .base import BaseSession

if TYPE_CHECKING:
    import aiosend
    from aiosend.client import APIServer
    from aiosend.methods import CryptoPayMethod
    from aiosend.types import _CryptoPayType


class AiohttpSession(BaseSession):
    """
    Http session based on aiohttp.

    This class is a wrapper of `aiohttp.ClientSession`.
    """

    def __init__(self, api_server: "APIServer") -> None:
        super().__init__(api_server)
        self._session: ClientSession | None = None

    async def request(
        self,
        token: str,
        client: "aiosend.CryptoPay",
        method: "CryptoPayMethod[_CryptoPayType]",
    ) -> "_CryptoPayType":
        """Make http request."""
        ssl_context = ssl.create_default_context(cafile=certifi.where())
        self._session = ClientSession(
            connector=TCPConnector(
                ssl_context=ssl_context,
            ),
        )
        async with self._session as session:
            resp = await session.post(
                url=self.api_server.url(method),
                data=method.model_dump_json(exclude_none=True),
                headers={
                    "Crypto-Pay-API-Token": token,
                    "Content-Type": "application/json",
                    "User-Agent": f"{SERVER_SOFTWARE} aiosend/{__version__}",
                },
            )
            response = self._check_response(client, method, await resp.text())
        return cast("_CryptoPayType", response.result)

    async def close(self) -> None:
        """Close http session."""
        if self._session is not None and not self._session.closed:
            await self._session.close()

    def __del__(self) -> None:
        """Close session connector."""
        if (
            self._session is not None
            and not self._session.closed
            and self._session.connector is not None
            and self._session.connector_owner
        ):
            self._session.connector.close()



from typing import TYPE_CHECKING

from aiosend.types import Invoice

if TYPE_CHECKING:
    import aiosend


class GetInvoice:
    """Get invoice."""

    async def get_invoice(
        self: "aiosend.CryptoPay",
        invoice: int | Invoice,
    ) -> Invoice | None:
        """
        Get exactly one invoice or none.

        Wrapper for :class:`aiosend.CryptoPay.get_invoices`

        Use this method to update status of an existing invoice
        object or to get this object by passing the invoice id.

        :return: :class:`aiosend.types.Invoice` object.
        """
        if isinstance(invoice, Invoice):
            invoice_id = invoice.invoice_id
        else:
            invoice_id = invoice

        invoices = await self.get_invoices(invoice_ids=[invoice_id])
        if invoices:
            return invoices[0]
        return None



from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Literal

    from aiosend.enums import Asset, Fiat
    from aiosend.methods import CryptoPayMethod


@dataclass(frozen=True)
class APIServer:
    """Configuration for endpoints."""

    name: str
    """Net name"""
    base: str
    """Base URL"""

    def url(self, method: "CryptoPayMethod") -> str:
        """Return URL for method."""
        return self.base.format(method=method.__method__)

    def get_qr(self, link: str) -> str:
        """Return qr code link."""
        return f"https://qr.crypt.bot/?url={link}"

    def get_image(
        self,
        asset: "Asset | str",
        asset_amount: float,
        fiat: "Fiat | str",
        fiat_amount: float,
        main: "Literal['asset', 'fiat']",
    ) -> str:
        """Return check image url."""
        return (
            "https://imggen.send.tg/checks/image?"
            f"asset={asset}"
            f"&asset_amount={asset_amount}"
            f"&fiat={fiat}"
            f"&fiat_amount={fiat_amount}"
            f"&main={main}"
        )


MAINNET = APIServer(
    name="MAINNET",
    base="https://pay.crypt.bot/api/{method}",
)
TESTNET = APIServer(
    name="TESTNET",
    base="https://testnet-pay.crypt.bot/api/{method}",
)



from .api_server import MAINNET, TESTNET, APIServer
from .client import CryptoPay
from .session.aiohttp import AiohttpSession
from .session.base import BaseSession

__all__ = (
    "MAINNET",
    "TESTNET",
    "APIServer",
    "AiohttpSession",
    "BaseSession",
    "CryptoPay",
)
